package net.sourceforge.noadsunpacker.cli;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import net.sourceforge.noadsunpacker.core.ArchiveUnpackManager;
import net.sourceforge.noadsunpacker.core.ArchiveUnpackManagerException;
import net.sourceforge.noadsunpacker.core.ArchiveUnpackManagerReportReceiver;
import net.sourceforge.noadsunpacker.gui.MainFrame;

/**
 * The main CLI class. Should be the main class
 * in the JAR file.
 * 
 * USAGE:
 * (without arguments)
 * -- Launches the GUI with blank fields
 *
 * noadsunpacker.jar <archive file> [target directory]
 * -- This launches the GUI with the specified <archive file> and <target directory>
 * -- If no directory is specified, uses the file's parent directory
 *
 * noadsunpacker.jar --cli <archive file> [target directory]
 * -- This unpacks the file into the specified directory without GUI
 * -- If no directory is specified, uses the file's parent directory
 *
 * noadsunpacker.jar --help
 * -- Prints help to standard output.
 *
 * noadsunpacker.jar --license
 * -- Prints the license to standard output.
 * EXIT CODES:
 * 0 - successful unpacking
 * 1 - incorrect usage
 * 100 - Error while unpacking
 *
 * @author Fedor Uvarov
 */
public class CLIMain {

public static final int EXIT_OK = 0;
public static final int EXIT_INCORRECT_USAGE = 1;
public static final int EXIT_EXCEPTION = 100;
public static final int EXIT_INIT_EXCEPTION = 200;
private static int ExitCode = 0;

public static void main(String[] args) {
    if (args.length == 0) {
        launchGUI("", "");
    }
    else if (args.length > 3) {
        if (args[0].equals("--cli")) {
            System.err.println("Incorrect usage. Run with --help to see the list of possible arguments.");
            System.err.println("Exiting");
            System.exit(EXIT_INCORRECT_USAGE);
        }
        else {
            System.err.println("Incorrect usage. Run with --help to see the list of possible arguments.");
            System.err.println("Launching GUI");
            launchGUI("", "");
        }
    }
    else {
        if (args[0].equals("--help")) {
            printFile("HELP");
            System.exit(EXIT_OK);
        }
        else if (args[0].equals("--license")) {
            printFile("NOTICE");
            System.out.println();
            System.out.println();
            System.out.println("THE CONTENTS OF THE FILE \"LICENSE\" FOLLOW");
            System.out.println();
            System.out.println();
            printFile("LICENSE");
            System.out.println();
            System.out.println();
            System.out.println("To read the file \"APACHE-LICENSE-2.0\" open it manually. If you received this program as a JAR file, unpack the JAR file with any ZIP archiver.");
        }
        else if (args[0].equals("--cli")) {
            try {
                if(args.length > 2) {
                    unpackInCLI(args[1], args[2]);
                }
                else unpackInCLI(args[1], ".");
            }
            catch(ArchiveUnpackManagerException e) {
                System.err.println("An error happened! More details below.");
                e.printStackTrace();
            }
        }
        else {
            if (args.length > 1) {
                launchGUI(args[0], args[1]);
            }
            else launchGUI(args[0], "");
        }
    }
}

private static void unpackInCLI(String FilePath, String TargetDirectory) throws ArchiveUnpackManagerException {
    ArchiveUnpackManager man = new ArchiveUnpackManager(FilePath, TargetDirectory, new ReportReceiver());
    man.unpack();
}

private static void launchGUI(String file, String dir) {
    if (!file.isEmpty() && dir.isEmpty()) {
        try {
            dir = new File(file).getCanonicalFile().getParent();
        }
        catch (IOException e) {
            dir = "";
        }
    }
    new MainFrame(file, dir).showFrame();
}

//To print help and license
private static void printFile(String whichOne) {
    try {
        InputStream f = ClassLoader.getSystemResource(whichOne).openStream();
        BufferedReader rd = new BufferedReader(new InputStreamReader(f, "UTF-8"));
        String s;
        while ((s = rd.readLine()) != null) {
            System.out.println(s);
        }
        rd.close();
    }
    catch (IOException e) {
        System.err.println("ERROR: an I/O error happened. More details below");
        e.printStackTrace();
    }
}

private static class ReportReceiver implements ArchiveUnpackManagerReportReceiver {
    public void report(int Report) {
        switch(Report) {
            case UNPACK_SUCCESS:
                System.exit(0);
            case UNPACK_FAILED:
                System.err.println("Unpacking failed for some reason");
                System.exit(1);
        }
    }

    public void report(int Report, Exception e) {
        if(Report == UNPACK_FAILED) {
            System.err.println("Unpacking failed. More details below.");
            e.printStackTrace();
            System.exit(1);
        }
    }

    public void report(int Report, String FilePath) {
        System.err.println("A file was unpacked but its size is incorrect. It may be corrupted. Path: "+FilePath);
    }

    public void reportUnpackingAbortedByUser() {
        throw new UnsupportedOperationException("Unexpected internal error");
    }

    public boolean verifyAbort() {
        throw new UnsupportedOperationException("Unexpected internal error");
    }

    public void reportTMFPUnpacked() {
        System.err.println("This is an uncompressed TMFP archive." +
                    "We cannot check its integrity yet. If you think this archive could be corrupted (e.g. while downloading), please check." +
                    "If it contains another archive, e.g. RAR, don't worry.");
    }
}
}
