package net.sourceforge.noadsunpacker.core;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

/**
 * Instances of this class are used to:
 * 1.Open an archive file
 * 2.Determine archive type
 * 3.Create a class which will perform the actual unpacking
 * 4.Start unpacking
 * 5.Open next parts when needed
 * 6.Report the actions during unpacking and results thereof
 *
 * @author Fedor Uvarov
 */
public class ArchiveUnpackManager implements UnpackManager {
public static final int ARCHIVE_TMFP = 1;
public static final int ARCHIVE_TMFP_UNPACKED = 2;
public static final int MEMORY_ALLOCATION_LIMIT = 10000000;    //10 Mb
private String SourceBase;
private int SourceIndex = 0;
private String RootOutputDirectory;
private int ArchiveType;
private File CurrentSourceInputFile;
private BufferedInputStream CurrentSourceInput;
private ArchiveUnpackManagerReportReceiver ReportReceiver;
private long BytesLeftInFile;
private UnpackManager UsedUnpackManager;
private boolean isSFX = false;
private volatile boolean shouldStop = false;
private volatile Thread BlockingThread; //This is the thread that blocks ability to continue

public ArchiveUnpackManager(String SourcePath, String RootOutputDirectory, ArchiveUnpackManagerReportReceiver ReportReceiver) throws ArchiveUnpackManagerException {
    //SAVING DATA
    this.ReportReceiver = ReportReceiver;
    CurrentSourceInputFile = new File(SourcePath);
    if(!CurrentSourceInputFile.exists())throw new ArchiveUnpackManagerException(ArchiveUnpackManagerException.REASON_FILENOTFOUND);
    this.SourceBase = CurrentSourceInputFile.getAbsolutePath();
    if(SourceBase.lastIndexOf(".") > SourceBase.lastIndexOf(File.separator))
        SourceBase = SourceBase.substring(0, SourceBase.lastIndexOf(".")+1);
    else SourceBase += ".";
    this.RootOutputDirectory = RootOutputDirectory;

    //OPENING THE FILE, CHECKING TYPE
    BytesLeftInFile = CurrentSourceInputFile.length();
    try {
        CurrentSourceInput = new BufferedInputStream(new FileInputStream(CurrentSourceInputFile));
    }
    catch(FileNotFoundException e) {
        throw new ArchiveUnpackManagerException(ArchiveUnpackManagerException.REASON_FILENOTFOUND, e);
    }
    try {
        byte[] tmpba = new byte[4];
        int tmpi = CurrentSourceInput.read(tmpba);
        if(tmpi != 4) {
            try {CurrentSourceInput.close();} catch(Exception e) {}
            throw new ArchiveUnpackManagerException(ArchiveUnpackManagerException.REASON_CANNOTREAD);
        }
        BytesLeftInFile -= 4;
        if(tmpba[0] == 0x50 && tmpba[1] == 0x53) {
            if(tmpba[2] == 0x43 && tmpba[3] == 0x46) ArchiveType = ARCHIVE_TMFP;
            else if(tmpba[2] == 0x50 && tmpba[3] == 0x46) ArchiveType = ARCHIVE_TMFP_UNPACKED;
            else throw new ArchiveUnpackManagerException(ArchiveUnpackManagerException.REASON_FORMATNOTSUPPORTED);
        }
        else if(tmpba[0] == 'M' && tmpba[1] == 'Z') {
            CurrentSourceInput.skip(0x3C - 4);  //Offset of the PE header offset. We move to 0x3C.
            BytesLeftInFile -= 0x3C - 4;

            //Read the PE header offset
            long tmpl = 0;
            for(int i = 0, multiplier = 1; i < 4; i++) {
                tmpi = CurrentSourceInput.read();
                if(tmpi == -1) throw new ArchiveUnpackManagerException(ArchiveUnpackManagerException.REASON_CANNOTREAD);
                tmpl += tmpi * multiplier;
                if (i < 3) multiplier *= 256;
            }
            BytesLeftInFile -= 4;
            //The header should be after this address
            if(tmpl < 0x3C) throw new ArchiveUnpackManagerException(ArchiveUnpackManagerException.REASON_FORMATNOTSUPPORTED);
            CurrentSourceInput.skip(tmpl - 0x3C - 4);   //We are at 0x3C + 4 (read bytes). We move to tmpl.
            BytesLeftInFile -= tmpl - 0x3C;
            CurrentSourceInput.read(tmpba);
            if (tmpba[0] == 'P' && tmpba[1] == 'E' && tmpba[2] == 0 && tmpba[3] == 0) {
                //It's an SFX archive. Now find the real header.
                isSFX = true;
                while(true) {
                    tmpi = CurrentSourceInput.read();
                    BytesLeftInFile--;
                    if(tmpi == 0x50) {
                        tmpi = CurrentSourceInput.read();
                        BytesLeftInFile--;
                        if(tmpi == 0x53) {
                            tmpi = CurrentSourceInput.read();
                            BytesLeftInFile--;
                            if(tmpi == 0x43) {
                                tmpi = CurrentSourceInput.read();
                                BytesLeftInFile--;
                                if(tmpi == 0x46) {   //Packed TMFP
                                    ArchiveType = ARCHIVE_TMFP;
                                    break;
                                }
                            }
                            if(tmpi == 0x50) {
                                tmpi = CurrentSourceInput.read();
                                BytesLeftInFile--;
                                if(tmpi == 0x46) {  //Unpacked TMFP
                                    ArchiveType = ARCHIVE_TMFP_UNPACKED;
                                    break;
                                }
                            }
                        }
                    }
                    if(tmpi == -1)
                        throw new ArchiveUnpackManagerException(ArchiveUnpackManagerException.REASON_FORMATNOTSUPPORTED);
                }
            }
        }
    }
    catch(IOException e) {
        try {CurrentSourceInput.close();} catch(Exception e2) {}
        throw new ArchiveUnpackManagerException(ArchiveUnpackManagerException.REASON_IOEXCEPTION, e);
    }

    //FILE OPENED, TYPE DETERMINED, CURRENT STREAM POSITION IS (magic number + 1)
    //Now we only have TMFP, so we create TMFPUnpackManager
    if(ArchiveType == ARCHIVE_TMFP_UNPACKED)ReportReceiver.reportTMFPUnpacked();
    UsedUnpackManager = new TMFPUnpackManager(this, ArchiveType != ARCHIVE_TMFP_UNPACKED);
}

public BufferedInputStream getCurrentSourceInput() {
    return CurrentSourceInput;
}

public String getRootOutputDirectory() {
    return RootOutputDirectory;
}

//Opens the next source file. Returns true if succeeds, false otherwise.
public boolean openNextSource() {
    try {
        //Closing current inpute
        CurrentSourceInput.close();
        
        //Generating the next name
        SourceIndex++;
        String NextSource;
        if(SourceIndex < 10)NextSource = SourceBase + "00";
        else if(SourceIndex < 100)NextSource = SourceBase + "0";
        else NextSource = SourceBase;
        NextSource += SourceIndex;

        //Opening the next file
        CurrentSourceInputFile = new File(NextSource);
        System.err.println(NextSource);
        if(!CurrentSourceInputFile.exists())return false;
        BytesLeftInFile = CurrentSourceInputFile.length();
        if(BytesLeftInFile == 0)return openNextSource();
        CurrentSourceInput = new BufferedInputStream(new FileInputStream(CurrentSourceInputFile));//Check the size

        //Returning successfully
        return true;
    }
    catch(IOException e) {
        return false;
    }
}

public long getBytesLeftInFile() {
    return BytesLeftInFile;
}

public File getCurrentSourceInputFile() {
    return CurrentSourceInputFile;
}

public void unpack() {
    UsedUnpackManager.unpack();
}

//The underlying UnpackManager reports that unpacking was aborted
public void reportUnpackingAborted(Exception e) {
    ReportReceiver.report(ArchiveUnpackManagerReportReceiver.UNPACK_FAILED, e);
}

//The underlying UnpackManager reports that unpacking was aborted by user
public void reportUnpackingAbortedByUser() {
    ReportReceiver.reportUnpackingAbortedByUser();
}

public void reportUnpackingSuccessful() {
    ReportReceiver.report(ArchiveUnpackManagerReportReceiver.UNPACK_SUCCESS);
}

public void reportOnePercentUnpacked() {
    ReportReceiver.report(ArchiveUnpackManagerReportReceiver.ONE_PERCENT_UNPACKED);
}

public void reportFileUnpackedIncorrectSize(File f) {
    ReportReceiver.report(ArchiveUnpackManagerReportReceiver.FILE_UNPACKED_SIZE_INCORRECT, f.getAbsolutePath());
}

public int getSourceIndex() {
    return SourceIndex;
}

public boolean isSFX() {
    return isSFX;
}

//Pauses unpacking process. Before the thread that called it dies,
//it should unpause(), otherwise the unpacking will abort
public void pause() {
    this.shouldStop = true;
    this.BlockingThread = Thread.currentThread();
}

public boolean shouldStop() {
    return shouldStop;
}

public boolean verifyAbort() {
    return ReportReceiver.verifyAbort();
}

public Thread getBlockingThread() {
    return BlockingThread;
}
}
