package net.sourceforge.noadsunpacker.core;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.zip.InflaterOutputStream;

/**
 * This class is used to unpack TMFP files (both
 * compressed and not compressed).
 *
 * @author Fedor Uvarov
 */
public class TMFPUnpackManager implements UnpackManager {

//Class representing a single entry in archive contents table
private class FileBlock {
    public boolean isDirectory; //Whether it is a directory
    public long Size;   //For files - size of corresponding block. For directories - how many of nthe following entries are within it.
    public long UnpackedSize = 0;
    public String Name; //Name
    public File CorrespondingFile = null;
}
public static final long PADDING_ALL = 2;    //Extra bytes at the end of every file
public static final long PADDING_SFX_EXTRA = 28; //Extra bytes at the end of an SFX archive (only the first volume)
public static final int MAXIMUM_MEMORY_ALLOCATION = 10000000; //Bytes to allocate for reading
//public static final long MEMORY_ALLOCATION_STEP = 10000; //Step when lowering the size
public static final int MINIMUM_MEMORY_ALLOCATION = 10000;
private ArrayList<FileBlock> FileTable;
private ArchiveUnpackManager ParentManager;
private boolean isPacked;
private BufferedInputStream CurrentSourceInput;
private long BytesLeftInFile;
private long UnpackedOnePercent = 0;
private long UnpackedLeftTillNextPercent;
private int NumberOfFiles;
private byte[] UsedBuffer;  //Buffer to read data

public TMFPUnpackManager(ArchiveUnpackManager ParentManager, boolean isPacked) {
    this.ParentManager = ParentManager;
    this.isPacked = isPacked;
}

//Unpacks current block of the source archive
protected void unpackBlock(File TargetFile, long NeededFileSize) throws UnpackException {
    //Reading the size
    long BlockSize = 0;
    int tmp;
    try {
        long multiplier = 1;
        //I believe they really do it like this
        for (int i = 0; i < 4; i++) {
            if (BytesLeftInFile == 0) openNextSource();
            tmp = CurrentSourceInput.read();
            if (tmp == -1)
                throw new UnpackException(UnpackException.REASON_UNPACKBLOCK_EOF, TargetFile.getPath());
            BytesLeftInFile--;
            tmp ^= 0x56;
            BlockSize += tmp * multiplier;
            if (i < 3) multiplier *= 256;
        }
        multiplier = 1;
        for (int i = 0; i < 4; i++) {
            if (BytesLeftInFile == 0) openNextSource();
            tmp = CurrentSourceInput.read();
            if (tmp == -1)
                throw new UnpackException(UnpackException.REASON_UNPACKBLOCK_EOF, TargetFile.getPath());
            BytesLeftInFile--;
            tmp ^= 0x56;
            BlockSize += tmp * multiplier;
            if (i < 3) multiplier *= 256;
        }
        //System.err.println("Block size: "+BlockSize);
    }
    catch (IOException e) {
        throw new UnpackException(UnpackException.REASON_UNPACKBLOCK_IOEXCEPTION, TargetFile.getPath(), e);
    }
    catch (UnpackException e) {
        if (e.getReason() == UnpackException.REASON_OPENNEXTSOURCE_FAILED)
            e.setFilePath(TargetFile.getPath());
        throw e;
    }

    //Creating output file
    OutputStream out;
    try {
        if (isPacked)
            out = new BufferedOutputStream(new InflaterOutputStream(new FileOutputStream(TargetFile)));
        else out = new BufferedOutputStream(new FileOutputStream(TargetFile));

    }
    catch (IOException e) {
        throw new UnpackException(UnpackException.REASON_UNPACKBLOCK_IOEXCEPTION, TargetFile.getPath(), e);
    }

    //Calculating size
    long OnePercentSize = (NeededFileSize == 0 ? 0 : NeededFileSize / 100);

    //Processing data
    try {
        continueUnpackBlock(out, TargetFile, BlockSize, OnePercentSize);
    }
    catch (UnpackException e) {
        try {
            out.close();
        }
        catch (IOException e2) {
        }
        e.setFilePath(TargetFile.getPath());
        throw e;
    }
    /*catch(IOException e) {
    try {out.close();} catch(IOException e2) {}
    throw new UnpackException(UnpackException.REASON_UNPACKBLOCK_IOEXCEPTION, TargetFile.getPath(), e);
    }*/

    //Closing
    try {
        out.close();
    }
    catch (IOException e) {
        throw new UnpackException(UnpackException.REASON_UNPACKBLOCK_IOEXCEPTION, TargetFile.getPath(), e);
    }
}

private void continueUnpackBlock(OutputStream out, File TargetFile, long BytesLeftToRead, long OnePercentSize) throws UnpackException {
    int BytesToReadNow, ReadBytesCount;
    long OldTargetFileSize = 0, NewTargetFileSize;
    try {
        while (true) {
            //Checking whether to abort
            if(ParentManager.shouldStop()) {
                if(ParentManager.verifyAbort()) {
                    throw new UnpackException(UnpackException.REASON_INTERNAL_UNPACK_ABORTED, null, null);
                }
            }
            if (BytesLeftInFile == 0) {
                openNextSource();
            }
            //Casts long -> int below should be safe
            if (BytesLeftToRead <= BytesLeftInFile)
                BytesToReadNow = (UsedBuffer.length < BytesLeftToRead ? UsedBuffer.length : (int) BytesLeftToRead);
            else
                BytesToReadNow = (UsedBuffer.length < BytesLeftInFile ? UsedBuffer.length : (int) BytesLeftInFile);
            ReadBytesCount = CurrentSourceInput.read(UsedBuffer, 0, BytesToReadNow);
            if (ReadBytesCount < BytesToReadNow)
                throw new UnpackException(UnpackException.REASON_CONTINUEUNPACKBLOCK_EOF, null);
            BytesLeftInFile -= BytesToReadNow;
            for (int i = 0; i < BytesToReadNow; i++)
                UsedBuffer[i] ^= 0x56;
            out.write(UsedBuffer, 0, BytesToReadNow);

            BytesLeftToRead -= BytesToReadNow;

            //Counting
            if(UnpackedOnePercent > 0) {
                NewTargetFileSize = TargetFile.length();
                UnpackedLeftTillNextPercent -= NewTargetFileSize - OldTargetFileSize;
                OldTargetFileSize = NewTargetFileSize;
                while(UnpackedLeftTillNextPercent <= 0) {
                    ParentManager.reportOnePercentUnpacked();
                    UnpackedLeftTillNextPercent += UnpackedOnePercent;
                }
            }
            
            if(BytesLeftToRead == 0)break;
            else if(BytesLeftToRead < 0) throw new UnpackException(UnpackException.REASON_CONTINUEUNPACKBLOCK_OVERREAD, null);
        }
    }
    catch (IOException e) {
        throw new UnpackException(UnpackException.REASON_CONTINUEUNPACKBLOCK_IOEXCEPTION, null, e);
    }
}

private void openNextSource() throws UnpackException {
    boolean res = ParentManager.openNextSource();
    if (!res) {
        throw new UnpackException(UnpackException.REASON_OPENNEXTSOURCE_FAILED, ParentManager.getCurrentSourceInputFile().getAbsolutePath());
    }
    BytesLeftInFile = ParentManager.getBytesLeftInFile() - PADDING_ALL;
    CurrentSourceInput = ParentManager.getCurrentSourceInput();
}

//Reads the table of contents
//As input it uses the descriptor of the temporary file where the unpacked table is stored
protected void readFileTable(File FileTableFile) throws UnpackException {
    int tmp;
    FileTable = new ArrayList<FileBlock>();
    FileBlock CurrentBlock;
    long multiplier;
    NumberOfFiles = 0;
    long TotalUnpackedSize = 0;

    //Creating decoder
    Charset NameCharset = Charset.forName("UTF-16LE");
    ByteArrayOutputStream bytebuf = new ByteArrayOutputStream();

    try {
        BufferedInputStream in = new BufferedInputStream(new FileInputStream(FileTableFile));
        while (true) {
            tmp = in.read();
            if (tmp == -1) break;
            CurrentBlock = new FileBlock();
            CurrentBlock.isDirectory = (((tmp - (tmp % 0x10)) / 0x10) % 2 != 0);
            in.skip(27); //Rest of attributes and all dates. TODO: implement preservation.

            //Reading size
            if (CurrentBlock.isDirectory) {
                multiplier = 1;
                for (int i = 0; i < 4; i++) {
                    tmp = in.read();
                    if (tmp == -1)
                        throw new UnpackException(UnpackException.REASON_READFILETABLE_EOF, FileTableFile.getAbsolutePath());
                    CurrentBlock.Size += tmp * multiplier;
                    if (i < 3) multiplier *= 256;
                }
            }
            else {
                //Reading size
                multiplier = 1;
                for (int i = 0; i < 8; i++) {
                    tmp = in.read();
                    if (tmp == -1)
                        throw new UnpackException(UnpackException.REASON_READFILETABLE_EOF, FileTableFile.getAbsolutePath());
                    CurrentBlock.UnpackedSize += tmp * multiplier;
                    if (i < 7) multiplier *= 256;
                }

                //Summing
                TotalUnpackedSize += CurrentBlock.UnpackedSize;

                /*multiplier = 256*256*256*256;
                for (int i = 0; i < 4; i++) {
                    tmp = in.read();
                    if (tmp == -1)
                        throw new UnpackException(UnpackException.REASON_READFILETABLE_EOF, FileTableFile.getAbsolutePath());
                    CurrentBlock.UnpackedSize += tmp*multiplier;
                    if (i < 3) multiplier *= 256;
                }*/
                NumberOfFiles++;
            }
            
            int bufint = 0;
            boolean secondint = false;
            while (true) {
                tmp = in.read();
                if (tmp == -1) {
                    in.close();
                    throw new UnpackException(UnpackException.REASON_READFILETABLE_EOF, FileTableFile.getAbsolutePath());
                }
                if (secondint) {
                    if (bufint == 0 && tmp == 0) break;
                    bytebuf.write(bufint);
                    bytebuf.write(tmp);
                }
                else bufint = tmp;
                secondint = !secondint;
            }
            CurrentBlock.Name = NameCharset.decode(ByteBuffer.wrap(bytebuf.toByteArray())).toString();

            //Adding block to the list
            FileTable.add(CurrentBlock);
            bytebuf.reset();
        }
    }
    catch (IOException e) {
        throw new UnpackException(UnpackException.REASON_READFILETABLE_IOEXCEPTION, FileTableFile.getAbsolutePath(), e);
    }

    UnpackedOnePercent = TotalUnpackedSize / 100;
    UnpackedLeftTillNextPercent = UnpackedOnePercent;
}

public void unpack() {
    //Getting data
    CurrentSourceInput = ParentManager.getCurrentSourceInput();
    BytesLeftInFile = ParentManager.getBytesLeftInFile() - PADDING_ALL;
    if(ParentManager.isSFX()) {
        BytesLeftInFile -= PADDING_SFX_EXTRA;
        if(!isPacked) BytesLeftInFile -= 1;
    }
    File RootDirectory = new File(ParentManager.getRootOutputDirectory());

    //Creating output directory
    if (!RootDirectory.exists()) RootDirectory.mkdirs();    //TODO: check for errors

    /* Allocating memory
     * In most cases we can determine the needed memory size judging by the first
     * file. Anyway, if we allocate too little, we'll unpack slowly
     * but nothing dangerous will happen.
     */
    int ArraySize;
    long memsize = Runtime.getRuntime().freeMemory();
    if (memsize < MINIMUM_MEMORY_ALLOCATION * 3) {
        //Throw an exception. Not enough memory.
    }
    if (MINIMUM_MEMORY_ALLOCATION > BytesLeftInFile)
        ArraySize = MINIMUM_MEMORY_ALLOCATION;
    else {
        if (BytesLeftInFile > MAXIMUM_MEMORY_ALLOCATION) {
            if (memsize > MAXIMUM_MEMORY_ALLOCATION)
                ArraySize = MAXIMUM_MEMORY_ALLOCATION;
            else ArraySize = (int) (memsize - MINIMUM_MEMORY_ALLOCATION * 2);
        }
        else {
            if (memsize > BytesLeftInFile)
                ArraySize = (int) BytesLeftInFile;
            else ArraySize = (int) (memsize - MINIMUM_MEMORY_ALLOCATION * 2);
        }
    }
    UsedBuffer = new byte[ArraySize];

    try {
        //Skipping the header
        CurrentSourceInput.skip(16);
        BytesLeftInFile -= 16;

        //Unpacking file table
        File FileTableFile = File.createTempFile("noads", "filetable");
        unpackBlock(FileTableFile, 0);
        readFileTable(FileTableFile);
        FileTableFile.delete();
        ParentManager.reportOnePercentUnpacked();

        //Unpacking files one by one
        FileBlock UsedMarkingFileBlock = new FileBlock();   //We'll use it instead of null
        FileBlock CurrentSubdirectory = UsedMarkingFileBlock;
        CurrentSubdirectory.CorrespondingFile = RootDirectory;
        CurrentSubdirectory.Size = 1;
        ArrayDeque<FileBlock> DirectoryStructure = new ArrayDeque<FileBlock>();
        File CurrentOutputFile;
        DirectoryStructure.add(UsedMarkingFileBlock);
        for (FileBlock file : FileTable) {
            if (file.isDirectory) {
                if (file.Size > 0) {
                    if(CurrentSubdirectory != UsedMarkingFileBlock) CurrentSubdirectory.Size--;
                    DirectoryStructure.addFirst(CurrentSubdirectory);
                    file.CorrespondingFile = new File(CurrentSubdirectory.CorrespondingFile.getAbsolutePath() + File.separator + file.Name);
                    CurrentSubdirectory = file;
                    CurrentSubdirectory.CorrespondingFile.mkdir();
                }
                else {
                    new File(CurrentSubdirectory.CorrespondingFile.getAbsolutePath() + File.separator + file.Name).mkdir();
                }
            }
            else {
                if (BytesLeftInFile == 0) openNextSource();
                CurrentOutputFile = new File(CurrentSubdirectory.CorrespondingFile.getAbsolutePath() + File.separator + file.Name);
                CurrentOutputFile.createNewFile();
                unpackBlock(CurrentOutputFile, file.UnpackedSize);
                if(CurrentOutputFile.length() != file.UnpackedSize) {
                    ParentManager.reportFileUnpackedIncorrectSize(file.CorrespondingFile);
                }
                if (CurrentSubdirectory != UsedMarkingFileBlock) {
                    CurrentSubdirectory.Size--;
                    if (CurrentSubdirectory.Size == 0) {
                        do {
                            CurrentSubdirectory = DirectoryStructure.pop();
                        } while (CurrentSubdirectory.Size == 0);
                    }
                }
            }
        }

        //Here we succeeded
        ParentManager.reportUnpackingSuccessful();
    }
    catch (IOException e) {
        ParentManager.reportUnpackingAborted(e);
    }
    catch (UnpackException e) {
        if(e.getReason() == UnpackException.REASON_INTERNAL_UNPACK_ABORTED) {
            ParentManager.reportUnpackingAbortedByUser();
        }
        else ParentManager.reportUnpackingAborted(e);
    }
}
}
